package ApplicationLayer.Circuit;

import ApplicationLayer.ApplicationLayer;
import ApplicationLayer.EmptyLoopException;
import ApplicationLayer.Equations;
import ApplicationLayer.IndependentCircuitsException;
import ApplicationLayer.NumOp;
import ApplicationLayer.Position;
import ApplicationLayer.ShortCircuitException;
import ApplicationLayer.Direction;
import PresentationLayer.Forms.MFState;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 *
 * @author Jakub Hrncir (2010)
 */
public class Circuit {

    public ArrayList<Node> nodes = new ArrayList<Node>(20);
    public ArrayList<Branch> branches = new ArrayList<Branch>(20);
    public ArrayList<Branch> tree = new ArrayList<Branch>(20);
    public ArrayList<Branch> outoftree = new ArrayList<Branch>();
    public ArrayList<Loop> loops = new ArrayList<Loop>();
    private ArrayList<ArrayList<Branch>> trees = new ArrayList<ArrayList<Branch>>();
    private ArrayList<ArrayList<Branch>> outOfTrees = new ArrayList<ArrayList<Branch>>();
    private ArrayList<ArrayList<Loop>> loopSystems = new ArrayList<ArrayList<Loop>>();

    /**
     * vytvoří prázdný obvod
     */
    public Circuit() {
    }

    public Circuit(ArrayList<Node> nodes, ArrayList<Branch> branches, ArrayList<Branch> tree, ArrayList<Branch> outoftree, ArrayList<Loop> loops) {
        this.nodes = nodes;
        this.branches = branches;
        this.tree = tree;
        this.outoftree = outoftree;
        this.loops = loops;
    }

    public boolean hasNode(Position pos) {
        Iterator<Node> it = nodes.iterator();
        while (it.hasNext()) {
            if (it.next().pos.is(pos)) {
                return true;
            }
        }
        return false;
    }

    public Node getNode(Position pos) {
        Iterator<Node> it = nodes.iterator();
        Node node;
        while (it.hasNext()) {
            node = it.next();
            if (node.pos.is(pos)) {
                return node;
            }
        }
        return null;
    }

    public boolean hasBranch(Node node, Direction dir) {
        Iterator<Branch> it = branches.iterator();
        Branch br;
        while (it.hasNext()) {
            br = it.next();
            if (br.nodes.get(0).pos.is(node.pos) & br.directions.get(0).ordinal() == dir.ordinal()) {
                return true;
            }
            if (br.nodes.get(1).pos.is(node.pos) & br.directions.get(1).ordinal() == dir.ordinal()) {
                return true;
            }
        }
        return false;
    }

    /**
     * @param node výchozí uzel
     * @param unmarked true/false: pouze neoznačené větve/všechny větve
     * @return větve vycházející z tohoto uzlu
     */
    public ArrayList<Branch> branches(Node node, boolean unmarked) {
        ArrayList<Branch> nextBr = new ArrayList<Branch>();
        Iterator<Branch> itBr = branches.iterator();
        while (itBr.hasNext()) {
            Branch br = itBr.next();
            if (br.hasNode(node)) {
                if (!unmarked | br.getMark() == 0) {
                    nextBr.add(br);
                }
            }
        }
        return nextBr;
    }

    /**
     * @return počet potřebných rovnic vyplývajících z 1. Kirchhoffova zákona
     */
    public int firsLawEquationsNeeded() {
        return (nodes.size() - 1);
    }

    /**
     * @return počet potřebných rovnic vyplývajících z 2. Kirchhoffova zákona
     */
    public int secondLawEquationsNeeded() {
        return (branches.size() - nodes.size() + 1);
    }

//------------------------------------------------------------------------------
    /**
     * nalezne různé úplné stromy el. sítě
     */
    private void findTrees() {
        trees.clear();
        Node node;
        //System.out.println("findTrees");
        for (int n = 0; n < nodes.size(); n++) {
            node = nodes.get(n);
            //System.out.println("\n node" + n);
            for (int b = 0; b < branches(node, false).size(); b++) {
                //System.out.println("  " + b);
                tree.clear();
                outoftree = (ArrayList<Branch>) branches.clone();
                firstTreeBranch(node, b);
                trees.add((ArrayList<Branch>) tree.clone());
                outOfTrees.add((ArrayList<Branch>) outoftree.clone());
            }
        }
        //System.out.println("Trees found:" + trees);
    }

    /**
     * začne vyhledávání stromu určiou větví (potřeba pro nalezení různých stromů)
     */
    private void firstTreeBranch(Node node, int index) {
        ArrayList<Branch> posible = branches(node, false);
        Branch br = posible.get(index);
        //System.out.println("   First branch: " + br);
        tree.add(br);
        outoftree.remove(br);
        //System.out.println("FindTreeBranch" + toString());
        findTreeBranch(br.otherNode(node));
    }

    /**
     * postupně přidává větve vedoucí k uzlům dosud neobsaženým ve stromu, rekurze pro nově připojené uzly
     * @param node výchozí uzel
     */
    private void findTreeBranch(Node node) {
        ArrayList<Branch> posible = branches(node, false);
        Iterator<Branch> it = posible.iterator();
        Branch br;
        while (it.hasNext()) {
            br = it.next();
            if (!treeHasNode(br.otherNode(node))) {
                tree.add(br);
                outoftree.remove(br);
                //System.out.println("FindTreeBranch" + toString());
                findTreeBranch(
                        br.otherNode(node));
            }
        }
    }

    /**
     * zjistí, zda-li je dotazovaný uzel obsažen ve stromu
     */
    private boolean treeHasNode(Node node) {
        Iterator<Branch> itBr = tree.iterator();
        Branch br;
        while (itBr.hasNext()) {
            br = itBr.next();
            if (br.hasNode(node)) {
                return true;
            }
        }
        return false;
    }

    /**
     * vyhledá stromy sítě, vygeneruje systémy smyček (pro každý strom) a poté vybere 1 systém 
     */
    public void generateLoopSystems() throws ShortCircuitException, EmptyLoopException {
        findTrees();
        loopSystems.clear();
        for (int i = 0; i < trees.size(); i++) {
            //System.out.println("generating loop system " + i);
            loops.clear();
            Iterator<Branch> it = outOfTrees.get(i).iterator();
            if (outOfTrees.get(i).size() != secondLawEquationsNeeded()) {
                //System.out.println("Number of outoftree branches doesn't match!");
            }
            while (it.hasNext()) {
                loops.add(new Loop(nodes, trees.get(i), it.next()));
                //System.out.println("New Loop"/* + toString()*/);
            }
            //System.out.println(loops);
            loopSystems.add((ArrayList<Loop>) loops.clone());
        }
        selectLoopSystem();
    }

    /**
     * vybere systém smyček s nejmenší složitostí soustavy rovnic
     */
    private void selectLoopSystem() throws ShortCircuitException {
        //System.out.println("Selecting loop System");
        int lastSize = Equations.getPredictedSize(new Circuit(nodes, branches, trees.get(0), outOfTrees.get(0), loopSystems.get(0)));
        int thisSize;
        tree = trees.get(0);
        outoftree = outOfTrees.get(0);
        loops = loopSystems.get(0);
        //System.out.println("size 0 : " + lastSize);
        for (int i = 1; i < trees.size(); i++) {
            thisSize = Equations.getPredictedSize(new Circuit(nodes, branches, trees.get(i), outOfTrees.get(i), loopSystems.get(i)));
            //System.out.println("size " + i + " : " + thisSize);
            if (thisSize < lastSize) {
                tree = trees.get(i);
                outoftree = outOfTrees.get(i);
                loops = loopSystems.get(i);
                lastSize = thisSize;
            }
        }
        //System.out.println("Selected size: " + lastSize);
    }

    /**
     * označí větve čísly
     * postupuje od uzlu A, pro každý uzel označí jeho zatím neoznačené větve po směru hodinových ručiček
     */
    public void markBranches() throws IndependentCircuitsException {
        orderNodes();
        Node node;
        Branch selected;
        Branch br;
        Iterator<Branch> it;
        Iterator<Node> itN = nodes.iterator();
        node = itN.next();
        //System.out.println("markBranches");
        //System.out.println(" " + node.toString());
        for (int i = 0; i < branches.size(); i++) {
            it = branches(node, true).iterator();
            //System.out.println(" " + node.toString() + " " + i);
            //System.out.println("  " + branches(node, true));
            try {
                selected = it.next();
            } catch (NoSuchElementException e) {
                throw new IndependentCircuitsException();
            }
            //System.out.println("  sel:" + selected.toString() + " has next " + it.hasNext());
            if (!it.hasNext()) {
                if (itN.hasNext()) {
                    node = itN.next();
                }
            }
            while (it.hasNext()) {
                br = it.next();
                if (br.getDirection(node).ordinal() < selected.getDirection(node).ordinal()) {
                    selected = br;
                }
            }
            selected.setMark(i + 1);
        }
        orderBranches();
    }

    /**
     * seřadí větve v arraylistu podle označení
     */
    private void orderBranches() {
        ArrayList<Branch> bs = (ArrayList<Branch>) branches.clone();
        branches.clear();
        for (int i = 1; i <= bs.size(); i++) {
            branches.add(getBranch(i, bs));
        }
    }

    /**
     * seřadí uzly v arraylistu podle označení
     */
    private void orderNodes() {
        ArrayList<Node> ns = (ArrayList<Node>) nodes.clone();
        nodes.clear();
        for (char c = 'A'; c <= 'A' + ns.size() - 1; c++) {
            nodes.add(getNode(c, ns));
        }
    }

    /**
     * @return větev s daným označením z daného arraylistu/null pokud tam taková není
     */
    private Branch getBranch(int hasMark, ArrayList<Branch> bs) {
        Iterator<Branch> it = bs.iterator();
        Branch b;
        while (it.hasNext()) {
            b = it.next();
            if (b.getMark() == hasMark) {
                return b;
            }
        }
        System.out.println("Error in Circuit:getBranch(int)");
        return null;
    }

    /**
     * @return uzel s daným označením z daného arraylistu/null pokud tam takový není
     */
    private Node getNode(char hasMark, ArrayList<Node> ns) {
        Iterator<Node> it = ns.iterator();
        Node n;
        while (it.hasNext()) {
            n = it.next();
            if (n.node.getCharMark() == hasMark) {
                return n;
            }
        }
        System.out.println("Error in Circuit:getBranch(int)");
        return null;
    }

//------------------------------------------------------------------------------
    /**
     * pro potřeby výpisů
     */
    @Override
    public String toString() {
        String s = "\nCircuit";
        Iterator<Node> itN = nodes.iterator();
        while (itN.hasNext()) {
            s = s.concat(itN.next().toString());
        }
        Iterator<Branch> itB = branches.iterator();
        while (itB.hasNext()) {
            s = s.concat(itB.next().toString());
        }
        s = s.concat("\nTree");
        itB = tree.iterator();
        while (itB.hasNext()) {
            s = s.concat(itB.next().toString());
        }
        s = s.concat("\nOut of Tree");
        itB = outoftree.iterator();
        while (itB.hasNext()) {
            s = s.concat(itB.next().toString());
        }
        Iterator<Loop> itL = loops.iterator();
        while (itL.hasNext()) {
            s = s.concat(itL.next().toString());
        }
        return (s + "\n--------");
    }

    /**
     * @return html string popis prvku(uzel,větev,smyčka zatím prázný) obvodu
     */
    public String getPartDescription(MFState state, Object part) {
        String s = "";
        s = ApplicationLayer.loadFileToString("html/header.html");
        if (part.getClass() == Node.class) {
            Node n = (Node) part;
            s = s.concat("<h3>Uzel " + n.node.getMark() + "</h3>");
            if (state == MFState.FORMING_EQUATIONS | state == MFState.EQUATIONS) {
                int branch;
                boolean dir;
                s += "<p>Podle zvolených směrů proudu:</p>";
                String sDo = "";
                String sZ = "";
                for (int i = 0; i < 4; i++) {
                    branch = n.node.getCurrent(Direction.values()[i]);
                    if (branch > 0) {
                        sDo += "<i>I</i><sub>" + n.node.branches[i] + "</sub><br>";
                    }
                    if (branch < 0) {
                        sZ += "<i>I</i><sub>" + n.node.branches[i] + "</sub><br>";
                    }
                }
                s += "<p>Do uzlu vchází proud/y</p><p class=\"tab\">" + sDo + "</p><p>Z uzlu odchází proud/y</p><p class=\"tab\"> " + sZ + "</p>";
            }
            if (state.ordinal()>2) {
                int branch;
                boolean dir;
                String sDo = "";
                String sZ = "";
                for (int i = 0; i < 4; i++) {
                    branch = n.node.getCurrent(Direction.values()[i]);
                    if (branch > 0) {
                        sDo += "<i>I</i><sub>" + n.node.branches[i] + "</sub> = " + NumOp.writeDouble(branches.get(n.node.branches[i] - 1).getI()) + "&nbsp;A<br>";
                    }
                    if (branch < 0) {
                        sZ += "<i>I</i><sub>" + n.node.branches[i] + "</sub> = " + NumOp.writeDouble(branches.get(n.node.branches[i] - 1).getI()) + "&nbsp;A<br>";
                    }
                }
                s += "<p>Do uzlu vchází proud/y</p><p class=\"tab\">" + sDo + "</p><p>Z uzlu odchází proud/y</p><p class=\"tab\"> " + sZ + "</p>";
            }
        }
        if (part.getClass() == Branch.class) {
            Branch b = (Branch) part;
            Iterator<Component> it = b.components.iterator();
            Component comp;
            String comps = "<h4>Komponenty</h4>";
            s += "<h3>Větev " + b.getMark() + "</h3>";
            if (state == MFState.FORMING_EQUATIONS | state == MFState.EQUATIONS) {
                s += "<p>Větví prochází proud <i>I</i><sub>" + b.getMark() + "</sub></p>";
                while (it.hasNext()) {
                    comp = it.next();
                    if (comp.isSource) {
                        comps += "<p><strong>Zdroj</strong> <i>U</i>&nbsp;=&nbsp;" + NumOp.writeDouble(comp.voltage) + "&nbsp;V</p>";
                    } else {
                        comps += "<p><strong>Rezistor</strong> <i>R</i>&nbsp;=&nbsp;" + NumOp.writeDouble(comp.resistance) + "&nbsp;Ω</p>";
                    }
                }
            }
            if (state.ordinal()>2) {
                s += "<p>Větví prochází proud <i>I</i><sub>" + b.getMark() + "</sub> = " + NumOp.writeDouble(b.getI()) + "&nbsp;A</p>";
                while (it.hasNext()) {
                    comp = it.next();
                    if (comp.isSource) {
                        comps += "<p><strong>Zdroj</strong> <i>U</i>&nbsp;=&nbsp;" + NumOp.writeDouble(comp.voltage) + "&nbsp;V</p>";
                    } else {
                        comps += "<p><strong>Rezistor</strong> <i>R</i>&nbsp;=&nbsp;" + NumOp.writeDouble(comp.resistance) + "&nbsp;Ω; <i>U</i>&nbsp;=&nbsp;" + NumOp.writeDouble(Math.abs(NumOp.roundDouble(comp.resistance * b.getI()))) + "&nbsp;V</p>";
                    }
                }
            }
            s += "<p>Celkový odpor větve je <i>R</i>&nbsp;=&nbsp;" + NumOp.writeDouble(b.getResistance()) + "&nbsp;Ω</p>";
            s += comps;
        }
        if (part.getClass() == Loop.class) {
            Loop l = (Loop) part;
        }
        return s;
    }

    /**
     * @return potřebuje některá část změnit směr v nákresu?
     */
    public boolean needInterpretSolution() {
        Iterator<Branch> it = branches.iterator();
        boolean need = false;
        while (it.hasNext()) {
            if (it.next().needInterpret())need = true;
        }
        return need;
    }

    /**
     * změní směry proudů v nákresu, aby odpovídaly kladým hodnotám proudů
     * odstraní směry proudů v nákresu, pokud proudy větví netečou
     */
    public void interpretSolution() {
        Iterator<Branch> it = branches.iterator();
        while (it.hasNext()) {
            it.next().interpretSolution();
        }
    }
}
